@@ -1,20 +1,21 @@
 # -*- coding: utf-8 -*-
+import keyword
 import os
 import re
 from pathlib import PurePosixPath, Path
-from typing import Union, Iterable, List, Dict
+from typing import Union, Iterable, List, Dict, Set
 
 import yaml
-from compose.config.types import ServicePort
-from ddb.feature import features
-from ddb.feature.traefik import TraefikExtraServicesAction
 from dotty_dict import Dotty
 from simpleeval import simple_eval
 
+from ddb.feature import features
+from ddb.feature.traefik import TraefikExtraServicesAction
 from .binaries import DockerBinary
+from .lib.compose.config.types import ServicePort
 from ...action import Action
 from ...action.action import EventBinding, InitializableAction
-from ...binary import binaries
+from ...binary import binaries, Binary
 from ...cache import caches, register_project_cache
 from ...config import config
 from ...context import context
@@ -137,6 +138,9 @@ class EmitDockerComposeConfigAction(Action):
             event_id = match.group(2)
             property_name = match.group(3)
 
+            if property_name and property_name in keyword.kwlist:
+                property_name = property_name + '_'
+
             args, kwargs = self._parse_value(value,
                                              {"service": service, "config": docker_compose_config},
                                              property_name)
@@ -187,7 +191,7 @@ class DockerComposeBinaryAction(InitializableAction):
 
     def __init__(self):
         super().__init__()
-        self.binaries = dict()
+        self.binaries = set()  # type: Set[Binary]
 
     @property
     def event_bindings(self):
@@ -204,38 +208,41 @@ class DockerComposeBinaryAction(InitializableAction):
 
     def destroy(self):
         if caches.has("docker.binaries"):
-            cache = caches.unregister("docker.binaries")
-            cache.close()
+            caches.unregister("docker.binaries", callback=lambda c: c.close())
 
     def before_events(self, docker_compose_config):
         """
         Reset current binaries dict.
         """
-        self.binaries = dict()
+        self.binaries = set()
 
     def after_events(self, docker_compose_config):
         """
-        Cache binaries and remove binaries from cache that are not found in current config.
+        Cache binaries and emit unregistered events for those removed from the previous run.
         """
         docker_binaries_cache = caches.get("docker.binaries")
+        cached_binaries = docker_binaries_cache.get("cached_binaries")  # type: Set[Binary]
+        if cached_binaries is None:
+            cached_binaries = set()
 
-        for name, binary in self.binaries.items():
-            docker_binaries_cache.set(name, binary)
+        to_remove_binaries = set()
+        for cached_binary in cached_binaries:
+            if cached_binary not in self.binaries:
+                if binaries.unregister(cached_binary.name, cached_binary):
+                    events.binary.unregistered(binary=cached_binary)
+                    to_remove_binaries.add(cached_binary)
 
-        binaries_to_remove = set()
-        for name in docker_binaries_cache.keys():
-            if name not in self.binaries.keys():
-                unregistered_binary = binaries.unregister(name)
-                events.binary.unregistered(binary=unregistered_binary)
-                binaries_to_remove.add(name)
+        for to_remove in to_remove_binaries:
+            cached_binaries.remove(to_remove)
 
-        for binary_to_remove in binaries_to_remove:
-            docker_binaries_cache.pop(binary_to_remove)
+        for binary in self.binaries:
+            cached_binaries.add(binary)
 
+        docker_binaries_cache.set("cached_binaries", cached_binaries)
         docker_binaries_cache.flush()
 
-    def execute(self, name=None, workdir=None, options=None, options_condition=None, args=None, exe=False,
-                docker_compose_service=None):
+    def execute(self, name=None, workdir=None, options=None, options_condition=None, condition=None, args=None,
+                exe=False, entrypoint=None, global_=None, docker_compose_service=None):
         """
         Execute action
         """
@@ -245,21 +252,17 @@ class DockerComposeBinaryAction(InitializableAction):
             raise ValueError("name should be defined")
 
         binary = DockerBinary(name, docker_compose_service=docker_compose_service, workdir=workdir, options=options,
-                              options_condition=options_condition, args=args, exe=exe)
+                              options_condition=options_condition, condition=condition, args=args, exe=exe,
+                              entrypoint=entrypoint, global_=global_)
 
-        self.binaries[name] = binary
+        self.binaries.add(binary)
 
-        if binaries.has(name):
-            existing_binary = binaries.get(name)
-            if existing_binary.is_same(binary):
-                context.log.notice("Binary exists: %s" % (name,))
-                events.binary.found(binary=binary)
-                return
-
-            binaries.unregister(name)
+        if binaries.has(name, binary):
+            context.log.notice("Binary exists: %s" % (name,))
+            events.binary.found(binary=binary)
+            return
 
         binaries.register(binary)
-
         context.log.success("Binary registered: %s", name)
         events.binary.registered(binary=binary)
 
@@ -289,6 +292,9 @@ class LocalVolumesAction(Action):
 
         volume_mappings = self._get_volume_mappings(docker_compose_config, external_volumes)
 
+        volume_mappings = [(source, target) for (source, target) in volume_mappings if
+                           self._check_file_in_project(source)]
+
         for source, target in volume_mappings:
             self._create_local_volume(source, target)
 
@@ -325,6 +331,12 @@ class LocalVolumesAction(Action):
                                       config.data.get("docker.user.uid"), config.data.get("docker.user.gid"),
                                       relative_path)
 
+    @staticmethod
+    def _check_file_in_project(target):
+        target_path = os.path.realpath(target)
+        cwd = os.path.realpath(".")
+        return target_path.startswith(cwd)
+
     @staticmethod
     def _create_local_volume(source, target):
         rel_source = os.path.relpath(source, ".")
@@ -332,7 +344,9 @@ class LocalVolumesAction(Action):
             context.log.notice("Local volume source: %s (exists)", rel_source)
         else:
             _, source_ext = os.path.splitext(source)
-            _, target_ext = os.path.splitext(target)
+            target_ext = None
+            if target:
+                _, target_ext = os.path.splitext(target)
             if source_ext or target_ext:
                 # Create empty file, because with have an extension in source or target.
                 os.makedirs(str(Path(source).parent), exist_ok=True)
@@ -349,7 +363,9 @@ class LocalVolumesAction(Action):
     @staticmethod
     def _get_volume_mappings(docker_compose_config, external_volumes):
         volume_mappings = []
-        for service in docker_compose_config['services'].values():
+        external_volumes_dict = {}
+
+        for service in docker_compose_config.get('services', {}).values():
             if 'volumes' not in service:
                 continue
 
@@ -361,12 +377,26 @@ class LocalVolumesAction(Action):
                     source, target, _ = volume_spec.rsplit(':', 2)
 
                 if source in external_volumes:
+                    external_volumes_dict[source] = target
                     continue
 
                 volume_mapping = (source, target)
 
                 if volume_mapping not in volume_mappings:
                     volume_mappings.append(volume_mapping)
+
+        for volume, volume_config in docker_compose_config.get('volumes', {}).items():
+            if volume_config and volume_config.get('driver') == 'local' and volume_config.get('driver_opts'):
+                driver_opts = volume_config.get('driver_opts')
+                device = driver_opts.get('device')
+                if device and driver_opts.get('o') == 'bind':
+                    source = PurePosixPath(device).joinpath(volume)
+                    target = external_volumes_dict.get(volume)
+                    if source and target:
+                        volume_mapping = (source, target)
+                        if volume_mapping not in volume_mappings:
+                            volume_mappings.append(volume_mapping)
+
         return volume_mappings
 
 
@@ -539,15 +569,15 @@ class DockerDisplayInfoAction(Action):
         :return: a dict containing useful labels data
         """
 
-        header = '{}'.format(service_name)
-        content = []
+        header_block = ['{}'.format(service_name)]
+        blocks = [header_block]
 
         if (config.args.type is None or 'env' in config.args.type) and environments:
             tmp_content = []
             for key in sorted(environments.keys()):
                 tmp_content.append('{}: {}'.format(key, environments.get(key)))
 
-            content.append(tmp_content)
+            blocks.append(tmp_content)
 
         if (config.args.type is None or 'port' in config.args.type) and ports:
             tmp_content = []
@@ -555,22 +585,22 @@ class DockerDisplayInfoAction(Action):
                 tmp_content.append(port.legacy_repr() if
                                    hasattr(port, 'legacy_repr') else
                                    "%s:%s" % (port.published, port.target))
-            content.append(tmp_content)
+            blocks.append(tmp_content)
 
         if (config.args.type is None or 'bin' in config.args.type) and docker_binaries:
             tmp_content = []
             for binary in sorted(docker_binaries):
                 tmp_content.append(binary)
-            content.append(tmp_content)
+            blocks.append(tmp_content)
 
         if (config.args.type is None or 'vhost' in config.args.type) and vhosts:
             tmp_content = []
             for vhost in sorted(vhosts):
                 tmp_content.append(vhost)
-            content.append(tmp_content)
+            blocks.append(tmp_content)
 
-        if len(content) > 0:
-            return get_table_display(header, content, False)
+        if blocks != [header_block]:
+            return get_table_display(blocks, False)
         return ''
 
     @staticmethod
@@ -584,7 +614,8 @@ class DockerDisplayInfoAction(Action):
         else:
             domain = extra_service_data.get('rule')
 
-        content = [
+        blocks = [
+            [id_ + " (extra)"],
             [domain + ' --> ' + extra_service_data.get('url')]
         ]
-        return get_table_display(id_ + " (extra)", content, False)
+        return get_table_display(blocks, False)
